@@ -17,12 +17,14 @@ use crate::{
     chunk_store::UsedSpace,
     chunks::Chunks,
     error::convert_to_error_message,
-    event_mapping::{map_routing_event, MsgContext},
+    event_mapping::{map_routing_event, Mapping, MsgContext},
     network::Network,
     node_ops::{MsgType, NodeDuty, OutgoingLazyError},
     state_db::{get_reward_pk, store_new_reward_keypair},
     Config, Error, Result,
 };
+use futures::{future::BoxFuture, lock::Mutex, stream::FuturesUnordered, FutureExt, StreamExt};
+use handle::NodeTask;
 use log::{error, warn};
 use rand::rngs::OsRng;
 use role::{AdultRole, Role};
@@ -31,11 +33,14 @@ use sn_messaging::client::ClientMsg;
 use sn_routing::{
     EventStream, {Prefix, XorName},
 };
+use std::sync::Arc;
 use std::{
     fmt::{self, Display, Formatter},
     net::SocketAddr,
     path::{Path, PathBuf},
 };
+use tokio::sync::RwLock;
+use tokio::task::JoinHandle;
 
 /// Static info about the node.
 #[derive(Clone)]
@@ -56,7 +61,6 @@ impl NodeInfo {
 /// Main node struct.
 pub struct Node {
     network_api: Network,
-    network_events: EventStream,
     node_info: NodeInfo,
     used_space: UsedSpace,
     role: Role,
@@ -64,7 +68,7 @@ pub struct Node {
 
 impl Node {
     /// Initialize a new node.
-    pub async fn new(config: &Config) -> Result<Self> {
+    pub async fn new(config: &Config) -> Result<(Self, EventStream)> {
         let root_dir_buf = config.root_dir()?;
         let root_dir = root_dir_buf.as_path();
         std::fs::create_dir_all(root_dir)?;
@@ -88,17 +92,22 @@ impl Node {
 
         let node = Self {
             role: Role::Adult(AdultRole {
-                chunks: Chunks::new(node_info.root_dir.as_path(), config.max_capacity()).await?,
+                chunks: Arc::new(RwLock::new(
+                    Chunks::new(node_info.root_dir.as_path(), config.max_capacity()).await?,
+                )),
             }),
             node_info,
             used_space: UsedSpace::new(config.max_capacity()),
-            network_api,
-            network_events,
+            network_api: network_api.clone(),
         };
 
-        messaging::send(node.register_wallet().await, &node.network_api).await?;
+        messaging::send(
+            Self::register_wallet(&network_api, reward_key).await,
+            &node.network_api,
+        )
+        .await?;
 
-        Ok(node)
+        Ok((node, network_events))
     }
 
     /// Returns our connection info.
@@ -116,33 +125,93 @@ impl Node {
         self.network_api.our_prefix().await
     }
 
+    async fn process_routing_event(
+        network_events: Arc<Mutex<EventStream>>,
+        network_api: Network,
+    ) -> Result<NodeTask> {
+        let node_task = if let Some(event) = network_events.lock().await.next().await {
+            let Mapping { op, ctx } = map_routing_event(event, &network_api).await;
+            NodeTask::Result(Box::new((vec![op], ctx)))
+        } else {
+            NodeTask::None
+        };
+        Ok(node_task)
+    }
+
     /// Starts the node, and runs the main event loop.
     /// Blocks until the node is terminated, which is done
     /// by client sending in a `Command` to free it.
-    pub async fn run(&mut self) -> Result<()> {
-        while let Some(event) = self.network_events.next().await {
-            // tokio spawn should only be needed around intensive tasks, ie sign/verify
-            let mapping = map_routing_event(event, &self.network_api).await;
-            self.process_while_any(mapping.op, mapping.ctx).await;
+    pub async fn run(&mut self, network_events: EventStream) -> Result<()> {
+        let network_api = self.network_api.clone();
+        let event_lock = Arc::new(Mutex::new(network_events));
+        let routing_task_handle = tokio::spawn(Self::process_routing_event(
+            event_lock.clone(),
+            network_api.clone(),
+        ));
+        let mut threads = FuturesUnordered::new();
+        threads.push(routing_task_handle);
+        while let Some(result) = threads.next().await {
+            match result {
+                Ok(Ok(NodeTask::Thread(handle))) => threads.push(handle),
+                Ok(Ok(NodeTask::Result(boxed))) => {
+                    let (duties, ctx) = *boxed;
+                    for duty in duties {
+                        let tasks = self.handle_and_get_threads(duty, ctx.clone()).await;
+                        threads.extend(tasks.into_iter());
+                    }
+                }
+                Ok(Ok(NodeTask::None)) => (),
+                Ok(Err(err)) => {
+                    let duty = try_handle_error(err, None);
+                    let tasks = self.handle_and_get_threads(duty, None).await;
+                    threads.extend(tasks.into_iter());
+                }
+                Err(err) => {
+                    error!("Error spawning task for task: {}", err);
+                }
+            }
+            // If the Mutex is locked, it means there is already a task running which
+            // is listening for routing events. If not, spawn a new task to listen for further events
+            if event_lock.try_lock().is_some() {
+                threads.push(tokio::spawn(Self::process_routing_event(
+                    event_lock.clone(),
+                    network_api.clone(),
+                )))
+            }
         }
-
         Ok(())
     }
 
-    // Keeps processing resulting node operations.
-    async fn process_while_any(&mut self, op: NodeDuty, ctx: Option<MsgContext>) {
-        let mut next_ops = vec![op];
-
-        while !next_ops.is_empty() {
-            let mut pending_node_ops: Vec<NodeDuty> = vec![];
-            for duty in next_ops {
-                match self.handle(duty).await {
-                    Ok(new_ops) => pending_node_ops.extend(new_ops),
-                    Err(e) => pending_node_ops.push(try_handle_error(e, ctx.clone())),
-                };
+    fn handle_and_get_threads(
+        &mut self,
+        op: NodeDuty,
+        ctx: Option<MsgContext>,
+    ) -> BoxFuture<Vec<JoinHandle<Result<NodeTask>>>> {
+        async move {
+            let mut threads = vec![];
+            match self.handle(op).await {
+                Ok(node_task) => match node_task {
+                    NodeTask::Result(boxed) => {
+                        let (node_duties, ctx) = *boxed;
+                        for duty in node_duties {
+                            let tasks = self.handle_and_get_threads(duty, ctx.clone()).await;
+                            threads.extend(tasks.into_iter());
+                        }
+                    }
+                    NodeTask::Thread(task_handle) => {
+                        threads.push(task_handle);
+                    }
+                    NodeTask::None => (),
+                },
+                Err(err) => {
+                    let duty = try_handle_error(err, ctx.clone());
+                    let tasks = self.handle_and_get_threads(duty, ctx.clone()).await;
+                    threads.extend(tasks.into_iter());
+                }
             }
-            next_ops = pending_node_ops;
+            threads
         }
+        .boxed()
     }
 }
 
